//============================================================================
// Name        : ehBinauralMultisourceTracker.cpp
// Author      : Eui-Hyun Kim (uihyun@gmail.com)
// Version     : 2.12.4
// Copyright   : Okuno Laboratory, Kyoto University
// Description :
//============================================================================

#include <iostream>
#include <string>
#include <list>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>


using namespace std;
using namespace HARK;

class BinauralMultisourceTracker;

DECLARE_NODE(BinauralMultisourceTracker);

/*Node
 *
 * @name BinauralMultisourceTracker
 * @category HARK:BINAURAL+
 * @description Multisource tracking by voice activity detection (VAD) and K-means clustering
 *
 * @input_name ESTIMATED_DOA
 * @input_type Vector<ObjectRef>
 * @input_description Estimated directions of multisource.
 *
 * @output_name TRACKED_DOA
 * @output_type Vector<ObjectRef>
 * @output_description Tracked directions of multisource.
 *
 * @parameter_name CLUSTERING_DURATION
 * @parameter_type float
 * @parameter_value 0.25
 * @parameter_description Time duration to cluster direction estimations [default: 0.25 seconds].
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The length in sample between a frame and a previous frame [default: 160].
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz) [default: 16000].
 *
 * END*/

class BinauralMultisourceTracker : public BufferedNode {

  int sourcesInputID;
  int spectrumInputID;
  int sourcesOutputID;
  int advance;
  int samplingRate;

  // k-means clustering
  float clusteringTimeDuration;
  int countClusteringTimeDuration;
  int startCount;
  list<Vector<ObjectRef> > listOfEstimatedDOAs;
  list<bool> listOfVADdicision;
  list<int> clusteringListOfEstimatedDOAs;

public:
  BinauralMultisourceTracker(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    sourcesInputID = addInput("ESTIMATED_DOA");
    sourcesOutputID = addOutput("TRACKED_DOA");
    advance = dereference_cast<int>(parameters.get("ADVANCE"));
    samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));

    inOrder = true;

    // k-means clustering
    clusteringTimeDuration = dereference_cast<float>(parameters.get("CLUSTERING_DURATION"));
    countClusteringTimeDuration = int(clusteringTimeDuration/float(advance)*float(samplingRate) + 0.5);

    inputsCache[sourcesInputID].lookBack = countClusteringTimeDuration;
    //outputs[sourcesOutputID].lookAhead = countClusteringTimeDuration;
  }

  virtual ~BinauralMultisourceTracker()
  {
  }

  void calculate(int output_id, int count, Buffer &out)
  {

    // ----------  Improved K-means Clustering for source tracking : Begin ----------

    // set input
    RCPtr<Vector<ObjectRef> > sourcesInputPtr = getInput(sourcesInputID, count+countClusteringTimeDuration/2);
    Vector<ObjectRef>& sourcesInput = *sourcesInputPtr;

    // mange data lists of estimated DOAs and VAD decision on each frame for clustering
    listOfEstimatedDOAs.push_back(sourcesInput);
    if (listOfEstimatedDOAs.size() > countClusteringTimeDuration)
      listOfEstimatedDOAs.pop_front();

    // store estimated DOAs for clustering
    clusteringListOfEstimatedDOAs.clear();
    for (int c = 0; c < countClusteringTimeDuration; c++) {

      list<Vector<ObjectRef> >::iterator iteratorForListOfEstimatedDOAs;
      for( iteratorForListOfEstimatedDOAs = listOfEstimatedDOAs.begin();
	   iteratorForListOfEstimatedDOAs != listOfEstimatedDOAs.end();
	   iteratorForListOfEstimatedDOAs++) {

	int nSources = iteratorForListOfEstimatedDOAs->size();

	for (int iSources = 0; iSources < nSources; iSources++)
	  {
	    RCPtr<Source> srcInput = (*iteratorForListOfEstimatedDOAs)[iSources];
	    int angle = (int)(180 / M_PI * atan2(srcInput->x[1], srcInput->x[0]));

	    clusteringListOfEstimatedDOAs.push_back(angle);
	  }

      }
    } // for (int c = 0; c < countClusteringTimeDuration; c++)

    // processing the improved K-means clustering
    int numberOfTrackedDOAs = 0;
    int trackedDOAs[60];
    int numberOfClusters = 1; // strat with one source
    int numberOfData = clusteringListOfEstimatedDOAs.size();
    int updatedCodeBook[60]; // assume the maximum number of clusters are 60.
    int previousCodeBook[60];

    int* data = new int[numberOfData];
    int** distances = new int*[numberOfData];
    int** minDistance = new int*[numberOfData];

    for (int i = 0; i < numberOfData; i++)
      minDistance[i] = new int[2]; // 0:data belong to 1:cluster

    list<int>::iterator iteratorForClusteringListOfEstimatedDOAs; int di;
    for(iteratorForClusteringListOfEstimatedDOAs = clusteringListOfEstimatedDOAs.begin(), di = 0; iteratorForClusteringListOfEstimatedDOAs != clusteringListOfEstimatedDOAs.end(); iteratorForClusteringListOfEstimatedDOAs++, di++)
      data[di] = (*iteratorForClusteringListOfEstimatedDOAs);


    while (numberOfData > countClusteringTimeDuration*2) {

      for (int i = 0; i < numberOfData; i++)
	distances[i] = new int[numberOfClusters];

      updatedCodeBook[numberOfClusters-1] = int(float(rand())*float(numberOfData)/RAND_MAX);
      previousCodeBook[numberOfClusters-1] = updatedCodeBook[numberOfClusters-1] + 1; // give differences from updatedCodeBook

      int DISTORTION = 0;
      bool sameCodeBook = false;
      while (!sameCodeBook) {

	// estimate distance between data and code
	for(int di = 0; di < numberOfData; di++)
	  for (int ci = 0; ci < numberOfClusters; ci++)
	    distances[di][ci] = abs(data[di] - updatedCodeBook[ci]);

	// estimate the minimum distance and its code
	for(int di = 0; di < numberOfData; di++) {

	  int min = distances[di][0];
	  minDistance[di][0] = distances[di][0];
	  minDistance[di][1] = 0;

	  for(int ci = 1; ci < numberOfClusters; ci++)
	    if (min > distances[di][ci]) {
	      minDistance[di][0] = distances[di][ci];
	      minDistance[di][1] = ci;
	      min = distances[di][ci];
	    }
	}

	// copy old CodeBook
	for (int ci = 0; ci < numberOfClusters; ci++)
	  previousCodeBook[ci] = updatedCodeBook[ci];

	// update CodeBook
	for (int ci = 0; ci < numberOfClusters; ci++) {

	  int meanCode = 0;
	  int averagingCount = 0;

	  for(int di = 0; di < numberOfData; di++)
	    if (ci == minDistance[di][1]) {
	      meanCode = meanCode + data[di];
	      averagingCount++;
	    }

	  if (averagingCount == 0) // it was error in choicing a center, so do again
	    updatedCodeBook[ci] = data[int(float(rand())*float(numberOfData)/RAND_MAX)];
	  else
	    updatedCodeBook[ci] = int(meanCode / averagingCount);
	}

	// check distortion
	DISTORTION = 0;
	for(int di = 0; di < numberOfData; di++)
	  DISTORTION = DISTORTION + minDistance[di][0];
	DISTORTION = DISTORTION / numberOfData;

	// check whether they are same or not
	sameCodeBook = true;
	for (int ci = 0; ci < numberOfClusters; ci++)
	  if (updatedCodeBook[ci] != previousCodeBook[ci]) {
	    sameCodeBook = false;
	    break;
	  }
      } // while (!sameCodeBook)

      // release memory
      for (int i = 0; i < numberOfData; i++)
	delete [] distances[i];

      // Increase step of the improved K-means clustering for tracking multisource
      if (DISTORTION > 3) {

	updatedCodeBook[numberOfClusters] = int(float(rand())*float(numberOfData)/RAND_MAX); // temporaly set a center of clusters
	previousCodeBook[numberOfClusters] = updatedCodeBook[numberOfClusters] + 1; // giving differences from updatedCodeBook
	numberOfClusters++;

      } else {

	// remove poor clusters and update codebook
	for (int ci = 0; ci < numberOfClusters; ci++) {
	  int CountingDataBelongToCluster = 0;
	  for(int di = 0; di < numberOfData; di++)
	    if (ci == minDistance[di][1])
	      CountingDataBelongToCluster++;

	  // Elimination step of the improved K-means clustering for tracking multisource // by 2013.12.5 *0.2f
	  if (float(CountingDataBelongToCluster) > float(numberOfData)*0.3f)
	    trackedDOAs[numberOfTrackedDOAs++] = updatedCodeBook[ci];
	}
	break; // escape while (clusteringListOfEstimatedDOAs.size() != 0)
      }
    } // while (numberOfData > countClusteringTimeDuration*2)

    // release memory
    for (int i = 0; i < numberOfData; i++)
      delete [] minDistance[i];

    delete [] distances;
    delete [] minDistance;
    delete [] data;

    // ----------  Improved K-means Clustering for source tracking : End ----------

    // Set output data
    RCPtr<Vector<ObjectRef> > sourcesOutputPtr(new Vector<ObjectRef>);
    Vector<ObjectRef>& sourcesOutput = *sourcesOutputPtr;
    (*outputs[sourcesOutputID].buffer)[count] = sourcesOutputPtr;

    for (int iSources = 0; iSources < numberOfTrackedDOAs; iSources++)
      {
	RCPtr<Source> srcOutput(new Source());
	srcOutput->power = 1.f; // Power of the direction localized
	srcOutput->remaining =  0.f; // An index that assumes how long the localized sound source lasts from then on
	srcOutput->id = iSources; // srcInput->id; still have a problem for multi-sound localization
	//srcOutput->id = 1; // srcInput->id; still have a problem for multi-sound localization
	float theta = trackedDOAs[iSources] * M_PI / 180.0f;  // angle
	// float theta = angle * M_PI / 180.0f;  // angle
	float phi = 0 * M_PI / 180.0f;   // elevation
	float cos_phi = cos(phi);
	srcOutput->x[0] = (float)(cos(theta) * cos_phi);
	srcOutput->x[1] = (float)(sin(theta) * cos_phi);
	srcOutput->x[2] = (float)(sin(phi));
	sourcesOutput.push_back(srcOutput);

	//	cout << "numberOfData" << numberOfData << "countClusteringTimeDuration" << countClusteringTimeDuration << endl;
	//	cout << "TRACKING:Frame#" << count << ", numberOfTrackedDOAs" << numberOfTrackedDOAs << ", trackedDOAs[iSources]"<< trackedDOAs[iSources]  << endl;
      }
  }
};

